/* source: xio-rawip.h */
/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
/* Published under the GNU General Public License V.2, see file COPYING */

#ifndef __xio_netlink_h_included
#define __xio_netlink_h_included 1

extern int xio_netlink_mtu(int interface_index, unsigned int mtu);

#endif /* !defined(__xio_netlink_h_included) */
